""" Release self updating cache, based on `cachetools.Cache` class. """

import asyncio
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union

import cachetools
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from tzlocal import get_localzone_name  # type: ignore

__all__ = ("SelfUpdatingCache",)

CacheKey = Union[int, str, float]
CacheValue = Union[Any, None]


class SelfUpdatingCache(cachetools.Cache):
    """
    My super-cache.
    Store requested objects while updater returns their value.

    Attributes:
        batchUpdater: batch updater function (List[key] -> List[Union[value, NO_RESULT]])
        scheduler: scheduler for updates
    """

    # an item removal flag. If updater returns this value, item will be removed.
    NO_RESULT: CacheValue = None
    MAX_PAGE_SIZE = 10000

    def __init__(
        self,
        maxsize,
        updateInterval,
        getsizeof=None,
        updater: Optional[Callable[[CacheKey], Awaitable[CacheValue]]] = None,
        batchUpdater: Optional[Callable[[List[CacheKey]], Awaitable[List[CacheValue]]]] = None,
    ):
        super().__init__(maxsize=maxsize, getsizeof=getsizeof)

        if updater is None and batchUpdater is None:
            raise ValueError(f"Use {cachetools.TTLCache} instead")
        if batchUpdater is None:
            if updater is None:
                raise ValueError("Use either 'updater' or 'batchUpdater'. Use 'batchUpdater' better.")

            async def _batchUpdater(keys: List[CacheKey]) -> List[CacheValue]:
                # pylint: disable-msg=E0102
                """
                Batch updater stub.

                Args:
                    keys: keys to update

                Returns:
                    values; SelfUpdatingCache.NO_RESULT - flag for item removal
                """
                results = await asyncio.gather(*[updater(key) for key in keys])  # type: ignore
                return results

            batchUpdater = _batchUpdater

        self.batchUpdater: Callable[[List[CacheKey]], Awaitable[List[CacheValue]]] = batchUpdater
        self.scheduler = AsyncIOScheduler(timezone=get_localzone_name())
        self.scheduler.start()
        self.scheduler.add_job(self._updateCache, trigger="interval", seconds=updateInterval)

    def getMany(self, keys: List[CacheKey]) -> List[CacheValue]:
        """
        Get multiple values.

        Args:
            keys: A list of keys to return values for

        Returns:
            A list with corresponding items from the cache. For every absent item None returned.
        """
        return [self.get(key) for key in keys]

    def setMany(self, update: Dict[CacheKey, CacheValue]):
        """
        Set multiple values. To remove particular items from the cache, set their values to NO_RESULT.

        Args:
            update: A dict with keys and values to set.

        """
        for key, value in update.items():
            if value is self.NO_RESULT:
                del self[key]
            else:
                self[key] = value

    async def _updateCache(self) -> None:
        """
        Update cache using self.batchUpdater.
        """
        keys = list(self)
        if not keys:
            return

        values = await self.batchUpdater(keys)
        self.setMany(dict(zip(keys, values)))

    async def close(self) -> None:
        """
        Close scheduler.
        """
        self.scheduler.shutdown()
