"""
Modules contains helper functions.
"""

import asyncio
import base64
import inspect
import platform
import re
from datetime import datetime, timedelta
from functools import wraps
from typing import Callable, Dict, List, Tuple
from uuid import UUID

from ciso8601 import parse_datetime
from dateutil import tz

from .regexps import UUID4_REGEXP, UUID_REGEXP

if platform.system() == "Windows":
    UTC_TIMEZONE = tz.tzutc()
    LOCAL_TIMEZONE = tz.tzlocal()
else:
    from udatetime import TZFixedOffset, to_string  # type: ignore # pylint: disable=import-error
    from udatetime._pure import local_utc_offset  # type: ignore # pylint: disable=import-error

    LOCAL_OFFSET_SECONDS = local_utc_offset * 60
    # udatetime.to_string works only with FixedOffset TZ
    UTC_TIMEZONE = TZFixedOffset(0)
    LOCAL_TIMEZONE = TZFixedOffset(int(local_utc_offset))


def isCoroutineFunction(func: Callable) -> bool:
    """
    Check function is coroutine or not.
    Args:
        func: some func

    Returns:
        true if  func is coroutine otherwise false
    """
    return getattr(func, "_is_coroutine", False) or inspect.iscoroutinefunction(func)


def isUUID4(uuidStr: str) -> bool:
    """
    Checking uuidStr is uuid4 in format '^[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$' or not

    Args:
        uuidStr: some string
    Returns:
        true if uuidStr is uuid4 else false
    >>> isUUID4("af3fb041-1a55-43e6-9ac2-a361aa432951")
    True
    >>> isUUID4("af3fb041-1a55-43e6-9ac2-a361aa43295")
    False
    >>> isUUID4("af3fb0411a5543e69ac2a361aa432951")
    False
    """
    match = UUID4_REGEXP.match(uuidStr)
    return bool(match)


def isUUID(uuidStr: str) -> bool:
    """
    Checking uuidStr is uuid4 in format '^[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$' or not

    Args:
        uuidStr: some string
    Returns:
        true if uuidStr is uuid4 else false
    >>> isUUID("af3fb041-1a55-w3e6-9ac2-a361aa432951")
    True
    >>> isUUID("af3fb041-1a55-43e6-9ac2-a361aa43295")
    False
    >>> isUUID("af3fb0411a5543e69ac2a361aa432951")
    False
    """
    match = UUID_REGEXP.match(uuidStr)
    return bool(match)


def nextUUID4(uuidStr: str) -> str:
    """
    Get next uuid4.

    Args:
        uuidStr: uuid4 as str

    Returns:
        next uuid4

    Raises:
        ValueError if uuid is not uuid4
        ValueError if uuid was exceed

    >>> nextUUID4('00000000-0000-4000-a000-000000000000')
    '00000000-0000-4000-a000-000000000001'
    >>> nextUUID4('00000000-0000-4fff-bfff-ffffffffffff')
    '00000000-0001-4000-8000-000000000000'
    """
    if uuidStr[14] != "4" or uuidStr[19] not in "89ab":
        raise ValueError(f"UUID is not UUID4: {uuidStr}")
    nextUID = UUID(int=UUID(uuidStr).int + 1)
    # make uuid4
    if nextUID.clock_seq_hi_variant // 16 == 0xC:
        # from '........-....-...0-C...-............'
        # make '........-....-...1-8...-............'
        nextUID = UUID(int=nextUID.int + (-0xC + 0x18) * 16**15)
        if nextUID.time_hi_version // 16**3 == 0x5:
            # from '........-...0-5...-....-............'
            # make '........-...1-4...-....-............'
            nextUID = UUID(int=nextUID.int + (-0x5 + 0x14) * 16**19)
    return str(nextUID)


def nextUUID(uuidStr: str) -> str:
    """
    Get next uuid4.

    Args:
        uuidStr: uuid as str

    Returns:
        next uuid4

    Raises:
        ValueError if uuid is not uuid4
        ValueError if uuid was exceed

    >>> nextUUID('00000000-0000-4000-a000-000000000000')
    '00000000-0000-4000-a000-000000000001'
    >>> nextUUID('00000000-0000-4fff-bfff-ffffffffffff')
    '00000000-0000-4fff-c000-000000000000'
    """
    uid = UUID(uuidStr)
    nextUID = UUID(int=uid.int + 1)
    return str(nextUID)


def getUUIDPages(
    count: int, gteId: str = str(UUID(int=0)), ltId: str = str(UUID(int=2**128 - 1))
) -> List[Tuple[str, str]]:
    """
    Split and get uuid range by disjoint pages

    Args:
        count: page count
        gteId: lower range bound
        ltId:  upper range bound

    Returns:
        list of pages. First element of tuple - lower page bound, second - lower upper bound
    """
    if count <= 1:
        return [(ltId, gteId)]
    rangeLength = UUID(ltId).int - UUID(gteId).int
    startValue = UUID(gteId).int
    endValue = UUID(ltId).int

    stepValue = rangeLength // count
    pages = []
    for pageIndex in range(count - 1):
        idLower = str(UUID(int=startValue + stepValue * (pageIndex)))
        idUpper = str(UUID(int=startValue + stepValue * (pageIndex + 1) - 1))
        pages.append((idLower, idUpper))

    pages.append((str(UUID(int=startValue + stepValue * (count - 1))), str(UUID(int=endValue))))
    return pages


def base64ToBytes(base64String: str) -> bytes:
    """
    Convert base64 string to bytes.

    Args:
        base64String: string

    Returns:
        bytes
    Raises:
        ValueError: if string does not valid base64 snting
    """
    return base64.b64decode(base64String.encode("utf-8"))


def bytesToBase64(binary: bytes) -> str:
    """
    Convert bytes to base64 string.

    Args:
        binary: some bytes

    Returns:
        string
    """
    return base64.b64encode(binary).decode("utf-8")


_FIRST_CAP_RE = re.compile("(.)([A-Z][a-z]+)")
_ALL_CAP_RE = re.compile("([a-z0-9])([A-Z])")


def convertToSnakeCase(camelCaseDict: Dict) -> dict:
    """
    Convert dict with camelCase keys to dict with snake_case keys
    Args:
        camelCaseDict: dict with keys in camelCase style

    Returns:
        dict with keys in snake_case style
    """
    res = {}
    for key, value in camelCaseDict.items():
        keyWitAfterFirstInsert = _FIRST_CAP_RE.sub(r"\1_\2", key)
        res[_ALL_CAP_RE.sub(r"\1_\2", keyWitAfterFirstInsert).lower()] = value
    return res


def convertStrFromSnakeToCamelCase(snakeCaseStr: str) -> str:
    """
    Convert string from snake to camel case
    Args:
        snakeCaseStr: string in camel case

    Returns:
        string in camel case
    """
    splitedStr = snakeCaseStr.split("_")
    res = "".join([splitedStr[0]] + [word.capitalize() for word in splitedStr[1:]])
    return res


def convertDictKeysFromSnakeToCamelCase(snakeCaseDict: dict) -> dict:
    """
    Convert string from snake to camel case
    Args:
        snakeCaseDict: dict with values in camel case

    Returns:
        dict with values in camel case
    """
    res = {convertStrFromSnakeToCamelCase(key): value for key, value in snakeCaseDict.items()}
    return res


def _convertTimeToStringUnix(inputDateTime: datetime, inUTC: bool) -> str:
    """
    Makes aware datetime and converts it to rfc3339 date time string.
    If inputDateTime is naive, its implied that its in UTC time zone.

    Args:
        inputDateTime: datetime object.
        inUTC: whether to adjust input datetime to UTC timezone
    Returns:
        string, format: "YYYY-MM-DDThh:mm:ss.sTZD" (eg 1997-07-16T19:20:30.45+01:00)"

    >>> _convertTimeToStringUnix(datetime(2018, 10, 12, 12, 0, tzinfo=tz.tzutc()), True)
    '2018-10-12T12:00:00.000000+00:00'

    >>> _convertTimeToStringUnix(datetime(2018, 10, 12, 12, 0, tzinfo=tz.tzutc()), False)
    '2018-10-12T15:00:00.000000+03:00'

    >>> _convertTimeToStringUnix(datetime(2018, 10, 12, 15, 0, tzinfo=tz.tzlocal()), True)
    '2018-10-12T12:00:00.000000+00:00'

    >>> _convertTimeToStringUnix(datetime(2018, 10, 12, 15, 0, tzinfo=tz.tzlocal()), False)
    '2018-10-12T15:00:00.000000+03:00'

    >>> _convertTimeToStringUnix(datetime(2018, 10, 12, 15, 0, tzinfo=None), True)
    '2018-10-12T15:00:00.000000+00:00'

    >>> _convertTimeToStringUnix(datetime(2018, 10, 12, 15, 0, tzinfo=None), False)
    '2018-10-12T18:00:00.000000+03:00'
    """

    if inputDateTime.tzinfo:
        if inUTC:
            convertedDateTime = inputDateTime.astimezone(UTC_TIMEZONE)
        else:
            convertedDateTime = inputDateTime.astimezone(LOCAL_TIMEZONE)
    elif not inputDateTime.tzinfo:
        if inUTC:
            convertedDateTime = inputDateTime
        else:
            convertedDateTime = inputDateTime.replace(tzinfo=LOCAL_TIMEZONE) + timedelta(seconds=LOCAL_OFFSET_SECONDS)

    return to_string(convertedDateTime)


def _convertTimeToStringWin(inputDateTime: datetime, inUTC: bool) -> str:
    """
    Function adds current time zone and isoformat(T) to datetime object
    Args:
        inputDateTime: datetime object with time zone info
        inUTC: return time in UTC format or local
    Raises:
        ValueError: if there is no time zone information
    Returns:
        string, format: "YYYY-MM-DDThh:mm:ss.sTZD" (eg 1997-07-16T19:20:30.45+01:00)"

    >>> _convertTimeToStringWin(datetime(2018, 10, 12, 12, 0, tzinfo=tz.tzutc()), True)
    '2018-10-12T12:00:00.000000+00:00'

    >>> _convertTimeToStringWin(datetime(2018, 10, 12, 12, 0, tzinfo=tz.tzutc()), False)
    '2018-10-12T15:00:00.000000+03:00'

    >>> _convertTimeToStringWin(datetime(2018, 10, 12, 15, 0, tzinfo=tz.tzlocal()), True)
    '2018-10-12T12:00:00.000000+00:00'

    >>> _convertTimeToStringWin(datetime(2018, 10, 12, 15, 0, tzinfo=tz.tzlocal()), False)
    '2018-10-12T15:00:00.000000+03:00'

    >>> _convertTimeToStringWin(datetime(2018, 10, 12, 15, 0, tzinfo=None), True)
    '2018-10-12T15:00:00.000000+00:00'

    >>> _convertTimeToStringWin(datetime(2018, 10, 12, 15, 0, tzinfo=None), False)
    '2018-10-12T18:00:00.000000+03:00'
    """
    if not inputDateTime.tzinfo:
        inputDateTime = inputDateTime.replace(tzinfo=UTC_TIMEZONE)

    outputTz = UTC_TIMEZONE if inUTC else LOCAL_TIMEZONE
    return inputDateTime.astimezone(tz=outputTz).isoformat(timespec="microseconds")


convertTimeToString = _convertTimeToStringUnix if platform.system() != "Windows" else _convertTimeToStringWin


def syncCallable(func: Callable) -> Callable:
    """
    Decorator for call asynchronous function in a synchronous code. Use keyword argument "asyncCall=False".

    Args:
        func: async function

    Returns:
        wrapped function
    """

    @wraps(func)
    def wrap(*funcArgs, asyncCall: bool = True, **funcKwargs):
        coroutine = func(*funcArgs, **funcKwargs)
        if asyncCall:
            return coroutine
        return asyncio.get_event_loop().run_until_complete(coroutine)

    return wrap


def convertToDateTimeObject(value: str) -> datetime:
    """
    Time will be converted to UTC without timezone information.

    Args:
        value: time string in isoformat "YYYY-MM-DDThh:mm:ss.sTZD" (eg 1997-07-16T19:20:30.45+01:00) or
               "YYYY-MM-DDThh:mm:ss.sZ"
    Returns:
        datetime object without timezone

    >>> convertToDateTimeObject("2020-01-01T12:00:00+03:00")
    datetime.datetime(2020, 1, 1, 9, 0)

    >>> convertToDateTimeObject("2020-01-01T12:00:00Z")
    datetime.datetime(2020, 1, 1, 12, 0)

    >>> convertToDateTimeObject("2020-01-01T12:00:00+00:00")
    datetime.datetime(2020, 1, 1, 12, 0)
    """
    datetimeValue = parse_datetime(value)
    return datetimeValue.astimezone(UTC_TIMEZONE).replace(tzinfo=None)
